"""
Manages the execution of restic command
"""
import datetime
import os
import re
import signal
from enum import Enum

from PyQt5.QtCore import QProcess, QProcessEnvironment

from enacrestic import const
from enacrestic.state import CurrentOperation, Operation, Status


class ResticCompletionStatus(Enum):
    """
    Enumerate all possible restic completion status
    """

    NO_ERROR = ""
    TIMEOUT = "timeout"
    REPO_LOCKED = "repo locked"


class ResticBackup:
    def __init__(self, app):
        self.app = app
        self._load_env_variables()
        self.current_utc_dt_starting = None
        self.p = None
        self.need_to_unlock = False

    def run(self):
        if not self.app.state.want_to_backup():
            self.app.logger.write_new_date_section(
                f"Backup not launched. "
                f"Current state is {self.app.state.current_operation.value}"
            )
            return

        # Run queued commands, one by one
        self._run_next_operation()

    def terminate(self):
        """
        Terminate currently running process, if any
        with SIGINT, equivalent to sending ctrl-c.
        This is the clean way to interrupt restic
        """
        if self.p is not None:
            os.kill(self.p.processId(), signal.SIGINT)

    def _run_next_operation(self):
        self.need_to_unlock = False
        next_operation = self.app.state.next_operation()
        self.app.qt_app.update_system_tray()
        if next_operation is None:
            return
        elif next_operation == Operation.PRE_BACKUP:
            self._run_prebackup()
        elif next_operation == Operation.BACKUP:
            self._run_backup()
        elif next_operation == Operation.FORGET:
            self._run_forget()
        elif next_operation == Operation.UNLOCK:
            self._run_unlock()

    def _load_env_variables(self):
        """
        Load expected env vars from ~/.enacrestic/env.sh
        to be used by restic commands
        """
        self.env = QProcessEnvironment.systemEnvironment()

        # Be sure to get output messages in english
        self.env.insert("LC_ALL", "C")

        variables_i_search = [
            r"RESTIC_\S+",
            r"AWS_ACCESS_KEY_ID",
            r"AWS_SECRET_ACCESS_KEY",
        ]
        try:
            with open(const.RESTIC_USER_PREFS["ENV"], "r") as f:
                for line in f.readlines():
                    # remove comments
                    # A) starting with #
                    # B) having ' #'
                    line = re.sub(r"^#.*", "", line)
                    line = re.sub(r"\s+#.*", "", line)

                    for var in variables_i_search:
                        match = re.match(r"export (%s)=(.*)$" % var, line)
                        if match:
                            self.env.insert(match.group(1), match.group(2))
        except FileNotFoundError:
            pass
        if not self.env.contains("RESTIC_REPOSITORY"):
            self.app.logger.error(
                f"{const.RESTIC_USER_PREFS['ENV']} seems not configured correctly"
            )

    def _run_prebackup(self):
        self.app.logger.write_new_date_section("Running restic pre_backup!")
        cmd = const.PRE_BACKUP_HOOK
        args = []
        self._run(cmd, args)

    def _run_backup(self):
        self.app.logger.write_new_date_section("Running restic backup!")
        cmd = "restic"
        args = [
            "backup",
            "--files-from",
            const.RESTIC_USER_PREFS["FILESFROM"],
            "--password-file",
            const.RESTIC_USER_PREFS["PASSWORDFILE"],
        ]
        if os.path.isfile(const.RESTIC_USER_PREFS["EXCLUDEFILE"]):
            args += ["--exclude-file", const.RESTIC_USER_PREFS["EXCLUDEFILE"]]
        self._run(cmd, args)

    def _run_forget(self):
        self.app.logger.write_new_date_section("Running restic forget!")
        cmd = "restic"
        args = [
            "forget",
            "--prune",
            "-g",
            "host",
            "-c",
            "--password-file",
            const.RESTIC_USER_PREFS["PASSWORDFILE"],
            "--keep-last",
            "3",
            "--keep-hourly",
            "24",
            "--keep-daily",
            "7",
            "--keep-weekly",
            "4",
            "--keep-monthly",
            "12",
            "--keep-yearly",
            "5",
        ]
        self._run(cmd, args)

    def _run_unlock(self):
        self.app.logger.write_new_date_section("Running restic unlock!")
        cmd = "restic"
        args = [
            "unlock",
            "--password-file",
            const.RESTIC_USER_PREFS["PASSWORDFILE"],
        ]
        self._run(cmd, args)

    def _run(self, cmd, args):
        self.p = QProcess()
        self.p.setProcessEnvironment(self.env)
        self.p.readyReadStandardOutput.connect(self._handle_stdout)
        self.p.readyReadStandardError.connect(self._handle_stderr)
        self.p.stateChanged.connect(self._handle_state)
        self.p.finished.connect(self._process_finished)
        self.p.start(cmd, args)
        self.current_process_completion_status = ResticCompletionStatus.NO_ERROR

    def _handle_stdout(self):
        data = self.p.readAllStandardOutput()
        stdout = bytes(data).decode("utf8")
        self.app.logger.write(stdout)

    def _handle_stderr(self):
        data = self.p.readAllStandardError()
        stderr = bytes(data).decode("utf8")
        self.app.logger.error(stderr)
        if self.app.state.current_operation in (
            CurrentOperation.BACKUP_IN_PROGRESS,
            CurrentOperation.FORGET_IN_PROGRESS,
            CurrentOperation.UNLOCK_IN_PROGRESS,
        ):
            if re.search(r"timeout", stderr):
                self.current_process_completion_status = ResticCompletionStatus.TIMEOUT
            if re.search(
                r"the `unlock` command can be used to remove stale locks", stderr
            ):
                self.current_process_completion_status = (
                    ResticCompletionStatus.REPO_LOCKED
                )
                lock_ago_str = re.search(
                    r"lock was created at.* \((.*)\)", stderr
                ).group(1)
                lock_hours_bool = re.search(r"\dh\d", lock_ago_str) is not None
                try:
                    lock_minutes_int = int(re.search(r"(\d+)m", lock_ago_str).group(1))
                except AttributeError:
                    lock_minutes_int = 0
                self.need_to_unlock = lock_hours_bool or lock_minutes_int > 30
                if not self.need_to_unlock:
                    self.app.logger.write(
                        "The lock is too new. We expect 30+ minutes to unlock manually."
                    )

    def _handle_state(self, proc_state):
        if proc_state == QProcess.Starting:
            self.current_utc_dt_starting = datetime.datetime.utcnow()
            self.app.qt_app.update_system_tray()
        elif proc_state == QProcess.NotRunning:
            self.current_chrono = (
                datetime.datetime.utcnow() - self.current_utc_dt_starting
            )

    def _process_finished(self):
        exitCode = self.p.exitCode()
        if self.p.exitStatus() == QProcess.NormalExit:
            if exitCode == 0:
                completion_status = Status.OK
            else:
                if (
                    self.current_process_completion_status
                    == ResticCompletionStatus.TIMEOUT
                ):
                    completion_status = Status.NO_NETWORK
                elif (
                    self.current_process_completion_status
                    == ResticCompletionStatus.REPO_LOCKED
                ):
                    completion_status = Status.REPO_LOCKED
                else:
                    completion_status = Status.LAST_OPERATION_FAILED
        else:
            completion_status = Status.LAST_OPERATION_FAILED
        self.app.state.finished_restic_cmd(
            completion_status,
            self.current_utc_dt_starting,
            self.current_chrono,
            self.need_to_unlock,
        )
        self.app.logger.write(
            f"Process finished ({exitCode}) in "
            f"{self.current_chrono.total_seconds():.2f} seconds "
            f"with status: '{completion_status.value}'\n\n"
        )

        self.current_utc_dt_starting = None
        self.p = None
        self._run_next_operation()



import datetime
import json
import os
from enum import Enum

from dynaconf import Dynaconf

from enacrestic import __version__, const
from enacrestic.utils import local_str_to_utc, utc_to_local_str


class Operation(Enum):
    """
    Enumerate all operations that app can run
    """

    PRE_BACKUP = "pre_backup"
    BACKUP = "backup"
    FORGET = "forget"
    UNLOCK = "unlock"


class CurrentOperation(Enum):
    """
    Enumerate all "what is the current operation" possible
    """

    JUST_LAUNCHED = "just_launched"
    PRE_BACKUP_IN_PROGRESS = "pre_backup_in_progress"
    BACKUP_IN_PROGRESS = "backup_in_progress"
    FORGET_IN_PROGRESS = "forget_in_progress"
    UNLOCK_IN_PROGRESS = "unlock_in_progress"
    IDLE = "idle"


class Status(Enum):
    """
    Enumerate all status possible (last command was ok or had an error)
    """

    OK = "ok"
    LAST_OPERATION_FAILED = "last_operation_failed"
    NO_NETWORK = "no_network"
    REPO_LOCKED = "repo_locked"


class State:
    """
    Load / Stores the state of the application
    """

    def __init__(self, app):
        self.app = app
        self.pre_backup_failed = False

    def __enter__(self):
        self._load()
        return self

    def __exit__(self, *args):
        self._save()

    def _load(self):
        conf_read = Dynaconf(
            settings_files=[const.RESTIC_STATEFILE],
        )
        self.current_operation = CurrentOperation.JUST_LAUNCHED
        self.current_status = Status.OK
        self.queue = []
        self.last_check_new_version_utc_dt = local_str_to_utc(
            conf_read.get("last_check_new_version_datetime", "1970-01-01 00:00:00")
        )
        self.latest_version_available = conf_read.get(
            "latest_version_available", __version__
        )
        self.nb_backups_before_forget = conf_read.get(
            "nb_backups_before_forget", self.app.conf.forget_every_n_backups
        )
        self.prev_backup_chronos = []
        for chrono in conf_read.get("prev_backup_chronos", []):
            self.prev_backup_chronos.append((local_str_to_utc(chrono[0]), chrono[1]))
        self.prev_forget_chronos = []
        for chrono in conf_read.get("prev_forget_chronos", []):
            self.prev_forget_chronos.append((local_str_to_utc(chrono[0]), chrono[1]))

    def _save(self):
        prev_backup_chronos = []
        for chrono in self.prev_backup_chronos:
            prev_backup_chronos.append((utc_to_local_str(chrono[0]), chrono[1]))
        prev_forget_chronos = []
        for chrono in self.prev_forget_chronos:
            prev_forget_chronos.append((utc_to_local_str(chrono[0]), chrono[1]))

        with open(const.RESTIC_STATEFILE, "w") as fh:
            json.dump(
                {
                    "current_operation": self.current_operation.value,
                    "current_status": self.current_status.value,
                    "last_check_new_version_datetime": utc_to_local_str(
                        self.last_check_new_version_utc_dt
                    ),
                    "latest_version_available": self.latest_version_available,
                    "nb_backups_before_forget": self.nb_backups_before_forget,
                    "prev_backup_chronos": prev_backup_chronos,
                    "prev_forget_chronos": prev_forget_chronos,
                    "version": __version__,
                },
                fh,
                sort_keys=True,
                indent=2,
            )

    def version_need_upgrade(self):
        """
        return bool if new version > current version
        """

        def try_to_int(string):
            try:
                return int(string)
            except ValueError:
                return string

        latest_version_info = tuple(
            try_to_int(ver) for ver in self.latest_version_available.split(".")
        )
        return latest_version_info > const.VERSION_INFO

    def get_icon(self):
        """
        return path to icon matching current state
        """
        if self.current_operation == CurrentOperation.JUST_LAUNCHED:
            return (
                f"{const.ICONS_FOLDER}/just_launched_badge.png"
                if self.version_need_upgrade()
                else f"{const.ICONS_FOLDER}/just_launched.png"
            )
        elif self.current_operation == CurrentOperation.IDLE:
            if self.current_status == Status.OK:
                if not self.pre_backup_failed:
                    return (
                        f"{const.ICONS_FOLDER}/backup_success_badge.png"
                        if self.version_need_upgrade()
                        else f"{const.ICONS_FOLDER}/backup_success.png"
                    )
                else:
                    return (
                        f"{const.ICONS_FOLDER}/error_badge.png"
                        if self.version_need_upgrade()
                        else f"{const.ICONS_FOLDER}/error.png"
                    )
            elif self.current_status == Status.LAST_OPERATION_FAILED:
                return (
                    f"{const.ICONS_FOLDER}/error_badge.png"
                    if self.version_need_upgrade()
                    else f"{const.ICONS_FOLDER}/error.png"
                )
            elif self.current_status == Status.NO_NETWORK:
                return (
                    f"{const.ICONS_FOLDER}/no_network_badge.png"
                    if self.version_need_upgrade()
                    else f"{const.ICONS_FOLDER}/no_network.png"
                )
            elif self.current_status == Status.REPO_LOCKED:
                return f"{const.ICONS_FOLDER}/repo_locked.png"
            else:
                self.app.logger.error(
                    f"unexpected state: {self.current_operation=} {self.current_status=}"
                )
                return (
                    f"{const.ICONS_FOLDER}/just_launched_badge.png"
                    if self.version_need_upgrade()
                    else f"{const.ICONS_FOLDER}/just_launched.png"
                )
        elif self.current_operation == CurrentOperation.PRE_BACKUP_IN_PROGRESS:
            return f"{const.ICONS_FOLDER}/pre_backup_in_progress.png"
        elif self.current_operation == CurrentOperation.BACKUP_IN_PROGRESS:
            return f"{const.ICONS_FOLDER}/backup_in_progress.png"
        elif self.current_operation == CurrentOperation.FORGET_IN_PROGRESS:
            return f"{const.ICONS_FOLDER}/forget_in_progress.png"
        elif self.current_operation == CurrentOperation.UNLOCK_IN_PROGRESS:
            return f"{const.ICONS_FOLDER}/unlock_in_progress.png"
        else:
            self.app.logger.error(
                f"unexpected state: {self.current_operation=} {self.current_status=}"
            )
            return (
                f"{const.ICONS_FOLDER}/just_launched_badge.png"
                if self.version_need_upgrade()
                else f"{const.ICONS_FOLDER}/just_launched.png"
            )

    def want_to_backup(self):
        """
        + Answer if a backup/forget can be run now
        + Set self.queue if possible
        """
        if self.current_operation in (
            CurrentOperation.IDLE,
            CurrentOperation.JUST_LAUNCHED,
        ):
            pre_backup_hook_is_executable = os.access(const.PRE_BACKUP_HOOK, os.X_OK)
            if pre_backup_hook_is_executable:
                self.queue = [Operation.PRE_BACKUP, Operation.BACKUP]
            else:
                self.queue = [Operation.BACKUP]
            return True
        else:
            return False

    def next_operation(self):
        """
        + return next Operation from self.queue
        + set self.current_operation accordingly
        + return None if nothing in the queue
        """
        if len(self.queue) > 0:
            operation = self.queue.pop(0)
            if operation == Operation.PRE_BACKUP:
                self.current_operation = CurrentOperation.PRE_BACKUP_IN_PROGRESS
                self.pre_backup_failed = False
            elif operation == Operation.BACKUP:
                self.current_operation = CurrentOperation.BACKUP_IN_PROGRESS
            elif operation == Operation.FORGET:
                self.current_operation = CurrentOperation.FORGET_IN_PROGRESS
            elif operation == Operation.UNLOCK:
                self.current_operation = CurrentOperation.UNLOCK_IN_PROGRESS
            else:
                self.app.logger.error(
                    f"unexpected Operation: {operation.value=} -> skipping"
                )
                return self.next_operation()

            return operation
        else:
            self.current_operation = CurrentOperation.IDLE
            return None

    def finished_restic_cmd(
        self, completion_status, start_utc_dt, chrono, queue_repo_unlock
    ):
        """
        + when success:
          + save chrono for current operation (backup or forget)
          + queue a forget if needed
        + otherwise:
          + empty queue
          + set self.last_failed_utc_dt
        """

        # Save Chrono if success
        if completion_status == Status.OK:
            chrono_seconds = round(chrono.total_seconds(), 2)  # Keep only 2 digits
            if self.current_operation == CurrentOperation.BACKUP_IN_PROGRESS:
                self.nb_backups_before_forget -= 1
                if self.nb_backups_before_forget <= 0:
                    self.nb_backups_before_forget = self.app.conf.forget_every_n_backups
                    self.queue.append(Operation.FORGET)
                self.prev_backup_chronos.insert(0, (start_utc_dt, chrono_seconds))
                if len(self.prev_backup_chronos) > const.NB_CHRONOS_TO_SAVE:
                    self.prev_backup_chronos.pop()
            elif self.current_operation == CurrentOperation.FORGET_IN_PROGRESS:
                self.prev_forget_chronos.insert(0, (start_utc_dt, chrono_seconds))
                if len(self.prev_forget_chronos) > const.NB_CHRONOS_TO_SAVE:
                    self.prev_forget_chronos.pop()
        elif completion_status == Status.REPO_LOCKED:
            self.last_failed_utc_dt = datetime.datetime.utcnow()
        else:
            if self.current_operation == CurrentOperation.PRE_BACKUP_IN_PROGRESS:
                self.pre_backup_failed = True
            else:
                # Don't do more things yet if NO_NETWORK or LAST_OPERATION_FAILED
                self.last_failed_utc_dt = datetime.datetime.utcnow()
                self.queue = []

        if queue_repo_unlock:
            if self.current_operation == CurrentOperation.BACKUP_IN_PROGRESS:
                self.queue.insert(0, Operation.BACKUP)
            elif self.current_operation == CurrentOperation.FORGET_IN_PROGRESS:
                self.queue.insert(0, Operation.FORGET)
            self.queue.insert(0, Operation.UNLOCK)

        self.current_status = completion_status

    def empty_queue(self):
        """
        Empty queue
        Useful when we want to quit the app and terminate processes in progress
        """
        self.queue = []



import datetime
import time

from enacrestic import const


def utc_to_local(utc_dt):
    """
    convert UTC datetime to local timezone datetime
    """
    epoch = time.mktime(utc_dt.timetuple())
    offset = datetime.datetime.fromtimestamp(
        epoch
    ) - datetime.datetime.utcfromtimestamp(epoch)
    return utc_dt + offset


def utc_to_local_str(utc_dt):
    """
    convert UTC datetime to local timezone datetime in string
    """
    return datetime.datetime.strftime(utc_to_local(utc_dt), const.DATE_FORMAT)


def local_to_utc(dt):
    """
    convert local timezone datetime to UTC datetime
    """
    epoch = time.mktime(dt.timetuple())
    offset = datetime.datetime.fromtimestamp(
        epoch
    ) - datetime.datetime.utcfromtimestamp(epoch)
    return dt - offset


def local_str_to_utc(dt):
    """
    convert local timezone datetime in string to UTC datetime
    """
    return local_to_utc(datetime.datetime.strptime(dt, const.DATE_FORMAT))



import json

from dynaconf import Dynaconf

from enacrestic import __version__, const


class Conf:
    """
    Load / Stores the config of the application
    """

    def __init__(self):
        pass

    def __enter__(self):
        self._load()
        return self

    def __exit__(self, *args):
        self._save()

    def _load(self):
        conf_read = Dynaconf(
            settings_files=[const.RESTIC_CONFFILE],
        )
        self.backup_every_n_minutes = conf_read.get(
            "backup_every_n_minutes", const.DEF_BACKUP_EVERY_N_MINUTES
        )
        self.forget_every_n_backups = conf_read.get(
            "forget_every_n_backups", const.DEF_FORGET_EVERY_N_BACKUPS
        )
        self.check_new_version_every_n_days = conf_read.get(
            "check_new_version_every_n_days", const.DEF_CHECK_NEW_VERSION_EVERY_N_DAYS
        )
        self.gui_autostart = conf_read.get("gui_autostart", const.DEF_GUI_AUTOSTART)

    def _save(self):
        """
        Save conf do file
        """
        with open(const.RESTIC_CONFFILE, "w") as fh:
            json.dump(
                {
                    "backup_every_n_minutes": self.backup_every_n_minutes,
                    "forget_every_n_backups": self.forget_every_n_backups,
                    "check_new_version_every_n_days": self.check_new_version_every_n_days,
                    "gui_autostart": self.gui_autostart,
                    "version": __version__,
                },
                fh,
                sort_keys=True,
                indent=2,
            )

    def set(self, **kwargs):
        """
        + Store new conf value
        + Save it to conf file
        """
        for key in (
            "backup_every_n_minutes",
            "forget_every_n_backups",
            "check_new_version_every_n_days",
            "gui_autostart",
        ):
            if kwargs.get(key) is not None:
                setattr(self, key, kwargs[key])
        self._save()



#!/usr/bin/env python3

import argparse

from enacrestic import __version__, app


def main():
    parser = argparse.ArgumentParser(
        prog="enacrestic", description="Automate restic execution"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--no-gui",
        dest="gui",
        action="store_const",
        const=False,
        default=True,
        help="disable GUI integration (system tray ++)",
    )
    args = parser.parse_args()

    app.App(gui_enabled=args.gui)


if __name__ == "__main__":
    main()



"""
Manages the writing to
+ log file
+ stdout

with auto log rotation (gz)
"""


import datetime
import gzip
import logging
import os
from logging.handlers import TimedRotatingFileHandler

from enacrestic import __version__, const


class Logger:
    class GZipRotator:
        def __call__(self, source, dest):
            os.rename(source, dest)
            with open(dest, "rb") as f_in:
                with gzip.open(f"{dest}.gz", "wb") as f_out:
                    f_out.writelines(f_in)
            os.remove(dest)

    def __init__(self, gui_enabled):
        self.logger = logging.getLogger()
        rotating_file_handler = TimedRotatingFileHandler(
            filename=const.RESTIC_LOGFILE,
            when="D",
            interval=const.LOGFILE_ROTATION_EVERY_N_DAYS,
            backupCount=const.LOGFILE_ROTATION_BACKUP_COUNT,
        )
        formatter = logging.Formatter("%(message)s")
        rotating_file_handler.setFormatter(formatter)
        rotating_file_handler.rotator = Logger.GZipRotator()
        self.logger.addHandler(rotating_file_handler)
        self.logger.setLevel(logging.INFO)
        if gui_enabled:
            self.app_flavor = ""
        else:
            self.app_flavor = "(noGUI) "

    def __enter__(self):
        self.write_new_date_section(
            f"Started ENACrestic {self.app_flavor}{__version__}\n"
        )
        return self

    def __exit__(self, typ, value, traceback):
        self.write_new_date_section(f"Stopped ENACrestic {__version__}\n")

    def write_new_date_section(self, message=None):
        section_header = "-" * 50 + f"\n{datetime.datetime.now()}"
        self.write(section_header)
        if message is not None:
            self.write(message)

    def write(self, message=""):
        self.logger.info(message)
        print(message)

    def error(self, message):
        lines = [f"! {line}" for line in message.split("\n")]
        self.logger.error("\n".join(lines))
        print("\n".join(lines))



#!/usr/bin/env python3

"""
Does :

+ Run every 30 minutes : `restic backup`
+ Run every 10 iteration : `restic forget` (for the backup rotation)

It uses the following files to be configured :

+ ~/.enacrestic/bkp_include
  This is given to --files-from option
+ ~/.enacrestic/bkp_exclude
  This is given to --exclude-file option
+ ~/.enacrestic/.pw
  This is given to --password-file option
+ ~/.enacrestic/env.sh
  This file has to configure env variables such as :
  + RESTIC_REPOSITORY
  + AWS_ACCESS_KEY_ID (if using S3)
  + AWS_SECRET_ACCESS_KEY (if using S3)

It writes to the following files :

+ ~/.enacrestic/last_backups.log
  All execution log
+ ~/.enacrestic/state.json
  State of the app (useful for upcoming executions)
+ ~/.enacrestic/prefs.json
  User preferences
"""

import datetime
import functools
import json
import os
import signal
import socket
import sys
import webbrowser

import requests
from pidfile import AlreadyRunningError, PIDFile
from PyQt5.QtCore import QCoreApplication, QTimer
from PyQt5.QtGui import QIcon
from PyQt5.QtNetwork import QAbstractSocket
from PyQt5.QtWidgets import QAction, QApplication, QMenu, QSystemTrayIcon

from enacrestic import __version__, const
from enacrestic.conf import Conf
from enacrestic.logger import Logger
from enacrestic.restic_backup import ResticBackup
from enacrestic.state import CurrentOperation, State, Status
from enacrestic.utils import utc_to_local


class QTNoGuiApp(QCoreApplication):
    """
    Main app, starting QCoreApplication
    used when --no-gui (typically on server)
    """

    def __init__(self, argv, app):
        super().__init__(argv)
        self.app = app

    def update_system_tray(self):
        pass  # no-gui


class QTGuiApp(QApplication):
    """
    Main app, starting QApplication and QSystemTrayIcon when it's ready.
    """

    def __init__(self, argv, app):
        super().__init__(argv)
        self.tray_icon = None
        self.app = app

        # start app when systray is available
        # workaround to fix automatic start when ENACrestic is launched at the session opening
        QTimer.singleShot(2000, self._start_app)

    def _start_app(self):
        """
        Start Qt System tray when everything is ready
        """
        icon_path = self.app.state.get_icon()
        self.tray_icon = QSystemTrayIcon(QIcon(icon_path), parent=self)
        self.tray_icon.show()

        menu = QMenu()
        # Entry to display informations to the user
        self.info_action = menu.addAction("ENACrestic launched")

        # Entry to display informations when an upgrade is available
        self.upgrade_action = menu.addAction(
            "New version is available.\nClick here to read the upgrade instructions."
        )
        self.upgrade_action.triggered.connect(self.open_upgrade_instructions)
        self.upgrade_action.setVisible(False)

        menu.addSection("Actions")

        # Entry to set if the application has
        # to auto-start with the session
        self.autostart_action = QAction("Auto-start", checkable=True)
        self.autostart_action.triggered.connect(self._toggle_autostart)
        self.autostart_action.setChecked(self.app.conf.gui_autostart)
        menu.addAction(self.autostart_action)

        # Entry to exit the application by the user
        exit_action = menu.addAction("Exit")
        exit_action.triggered.connect(self.app.quit)

        self.tray_icon.setContextMenu(menu)

        self.update_system_tray()

    def open_upgrade_instructions(self):
        webbrowser.open(const.UPGRADE_DOC)

    def update_system_tray(self):
        """
        update system tray according to the state
        + icon to current state
        + info_action with current state infos
        + Show upgrade_action if needed
        """

        def _str_date(utc_dt):
            """
            return nice date (with only h:m:s if it's in the last 24h)
            """

            if datetime.datetime.utcnow() - utc_dt < datetime.timedelta(days=1):
                return "at %s" % utc_to_local(utc_dt).strftime("%H:%M:%S")
            else:
                return "on %s" % utc_to_local(utc_dt).strftime("%Y-%m-%d %H:%M:%S")

        def _str_duration(seconds, shortest=False):
            """
            return nice duration as __h __m __s
            if not shortest :
                __s | __m __s | __h __m __s
            if shortest :
                __h | __m | __s |
                __h __m | __m __s |
                __h __m __s
            """
            seconds = int(seconds)
            hours, seconds = divmod(seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            if shortest:
                if hours > 0:
                    if minutes > 0:
                        if seconds > 0:
                            return f"{hours}h {minutes}m {seconds}s"
                        else:
                            return f"{hours}h {minutes}m"
                    else:
                        if seconds > 0:
                            return f"{hours}h {minutes}m {seconds}s"
                        else:
                            return f"{hours}h"
                else:
                    if minutes > 0:
                        if seconds > 0:
                            return f"{minutes}m {seconds}s"
                        else:
                            return f"{minutes}m"
                    else:
                        return f"{seconds}s"
            else:
                if hours > 0:
                    return f"{hours}h {minutes}m {seconds}s"
                elif minutes > 0:
                    return f"{minutes}m {seconds}s"
                else:
                    return f"{seconds}s"

        def _str_last_chronos(subject, list_chronos):
            """
            return msg with latest chrono and average over the last n
            """
            nb_chronos = len(list_chronos)
            if nb_chronos == 0:
                return ""
            msg = """
-> latest %s %s : %s""" % (
                subject,
                _str_date(list_chronos[0][0]),
                _str_duration(list_chronos[0][1]),
            )
            if nb_chronos >= 2:
                sum_chronos = sum([chrono[1] for chrono in list_chronos])
                average_chrono = sum_chronos / nb_chronos
                msg += """
average over the last %d : %s""" % (
                    nb_chronos,
                    _str_duration(average_chrono),
                )
            return msg

        if self.tray_icon is None:
            return

        icon_path = self.app.state.get_icon()
        self.tray_icon.setIcon(QIcon(icon_path))

        state_msg = f"ENACrestic {__version__}\n\n"

        if self.app.state.current_operation == CurrentOperation.JUST_LAUNCHED:
            state_msg += (
                "Just launched,\n"
                "a backup will be done every "
                f"{_str_duration(self.app.conf.backup_every_n_minutes * 60, True)}."
            )
        elif self.app.state.current_operation == CurrentOperation.IDLE:
            if self.app.state.current_status == Status.OK:
                state_msg += "Last backup was successful"
                if self.app.state.pre_backup_failed:
                    state_msg += " but pre-backup hook failed"
            elif self.app.state.current_status == Status.LAST_OPERATION_FAILED:
                state_msg += (
                    f"Last operation failed, {_str_date(self.app.state.last_failed_utc_dt)}\n"
                    f"see {const.RESTIC_LOGFILE} for details."
                )
            if self.app.state.current_status == Status.NO_NETWORK:
                state_msg += (
                    f"Network timeout {_str_date(self.app.state.last_failed_utc_dt)}"
                )
        elif (
            self.app.state.current_operation == CurrentOperation.PRE_BACKUP_IN_PROGRESS
        ):
            state_msg += "Pre-backup in progress"
            if self.app.restic_backup.current_utc_dt_starting is not None:
                state_msg += f" (started {_str_date(self.app.restic_backup.current_utc_dt_starting)})"
        elif self.app.state.current_operation == CurrentOperation.BACKUP_IN_PROGRESS:
            state_msg += "Backup in progress"
            if self.app.restic_backup.current_utc_dt_starting is not None:
                state_msg += f" (started {_str_date(self.app.restic_backup.current_utc_dt_starting)})"
        elif self.app.state.current_operation == CurrentOperation.FORGET_IN_PROGRESS:
            state_msg += "Cleanup in progress"
            if self.app.restic_backup.current_utc_dt_starting is not None:
                state_msg += f" (started {_str_date(self.app.restic_backup.current_utc_dt_starting)})"
        elif self.app.state.current_operation == CurrentOperation.UNLOCK_IN_PROGRESS:
            state_msg += "Unlock in progress"
            if self.app.restic_backup.current_utc_dt_starting is not None:
                state_msg += f" (started {_str_date(self.app.restic_backup.current_utc_dt_starting)})"

        # Add conditionnal stats on last backups and last cleanups
        last_chronos = _str_last_chronos("backup", self.app.state.prev_backup_chronos)
        last_chronos += _str_last_chronos("cleanup", self.app.state.prev_forget_chronos)
        if last_chronos != "":
            state_msg += "\n"
            state_msg += last_chronos
        self.info_action.setText(state_msg)

        # show / hide upgrade_action
        self.upgrade_action.setVisible(self.app.state.version_need_upgrade())

    def _toggle_autostart(self):
        """
        Save and apply user's choice to autostart or not
        """

        gui_autostart = self.autostart_action.isChecked()
        self.app.conf.set(gui_autostart=gui_autostart)
        if gui_autostart:
            # Want the app to autostart with user's session
            autostart_folder = os.path.dirname(const.RESTIC_AUTOSTART_FILE)
            os.makedirs(autostart_folder, exist_ok=True)
            with open(const.RESTIC_AUTOSTART_FILE, "w") as f:
                f.write(
                    f"""\
[Desktop Entry]
Name=ENACrestic
Comment=Automated Backup with restic
Exec={const.ENACRESTIC_BIN}
Icon=enacrestic
Terminal=false
Type=Application
Encoding=UTF-8
Categories=Utility;Archiving;
Keywords=backup;enac;restic
Name[en_US]=ENACrestic
X-GNOME-Autostart-enabled=true
"""
                )
        else:
            # Users doesn't want ENACrestic to autostart
            try:
                os.remove(const.RESTIC_AUTOSTART_FILE)
            except FileNotFoundError:
                pass


class SignalWatchdog(QAbstractSocket):
    """
    Watchdog to propagates system signals from Python to QEventLoop
    https://stackoverflow.com/a/65802260/446302
    This is necessary to handle SIGINT signals
    """

    def __init__(self):
        super().__init__(QAbstractSocket.SctpSocket, None)
        self.writer, self.reader = socket.socketpair()
        self.writer.setblocking(False)
        signal.set_wakeup_fd(self.writer.fileno())  # Python hook
        self.setSocketDescriptor(self.reader.fileno())  # Qt hook
        self.readyRead.connect(lambda: None)  # Dummy function call


def sigterm_handler(app, signal_num, stack_frame):
    """
    Handler for the SIGINT signal
    Will ensures that everything is closed on SIGTERM
    """
    app.quit()


class App:
    def __init__(self, gui_enabled=True):
        self.gui_enabled = gui_enabled
        # Create pref folder if doesn't exist yet
        if not os.path.exists(const.ENACRESTIC_PREF_FOLDER):
            os.makedirs(const.ENACRESTIC_PREF_FOLDER)

        with Logger(self.gui_enabled) as self.logger:
            try:
                with PIDFile(const.PID_FILE):
                    with Conf() as self.conf:
                        with State(self) as self.state:
                            self.restic_backup = ResticBackup(self)
                            self._start_app()
                            sys.exit(self.qt_app.exec_())
            except AlreadyRunningError:
                self.logger.write("Already running -> quit")

    def _start_app(self):
        """
        + Start GUI/noGUI Qt
        + manages SIGTERM
        + Launch timers that will trigger expected operations
        """
        if self.gui_enabled:
            self.qt_app = QTGuiApp(sys.argv, self)
        else:
            self.qt_app = QTNoGuiApp(sys.argv, self)
        signal.signal(
            signal.SIGTERM,
            functools.partial(
                sigterm_handler,
                self,
            ),
        )

        self.next_backup_timer = QTimer()
        self.next_backup_timer.timeout.connect(self.restic_backup.run)
        self.next_backup_timer.start(self.conf.backup_every_n_minutes * 60_000)

        self.check_for_latest_version_timer = QTimer()
        self.check_for_latest_version_timer.timeout.connect(
            self._maybe_check_for_latest_version
        )
        self.check_for_latest_version_timer.start(86_400_000)  # every hour
        self._maybe_check_for_latest_version()

        self.signal_watchdog = SignalWatchdog()

    def _maybe_check_for_latest_version(self):
        """
        If enough time has passed since last check
        + Check for latest version on PYPI and store it in self.state.latest_version_available (as str)
        + Update self.state.last_check_new_version_utc_dt with curent datetime
        + It writes info to logger about it.
        """
        next_check_utc_dt = (
            self.state.last_check_new_version_utc_dt
            + datetime.timedelta(days=self.conf.check_new_version_every_n_days)
        )
        if datetime.datetime.utcnow() > next_check_utc_dt:
            try:
                self.logger.write_new_date_section("Checking for latest release")
                pypi_response = requests.get(const.PYPI_PROJECT_URL)
                pypi_json = pypi_response.json()
                self.state.latest_version_available = pypi_json["info"]["version"]
                if self.state.version_need_upgrade():
                    self.logger.write(
                        f"new release available : {self.state.latest_version_available}"
                    )
                else:
                    self.logger.write("ok")
                self.state.last_check_new_version_utc_dt = datetime.datetime.utcnow()
            except (
                requests.exceptions.ConnectionError,
                json.decoder.JSONDecodeError,
                KeyError,
            ):
                self.logger.error(
                    "Could not retrieve latest release number. "
                    "Considering it's fine."
                )
                self.state.latest_version_available = __version__
            self.qt_app.update_system_tray()

    def quit(self):
        """
        triggered when the app is being closed
        """
        self.state.empty_queue()
        if self.state.current_operation in (
            CurrentOperation.BACKUP_IN_PROGRESS,
            CurrentOperation.FORGET_IN_PROGRESS,
        ):
            self.logger.write(
                "Closing the app. Waiting for restic process to be finished"
            )
            self.restic_backup.terminate()
        QTimer.singleShot(100, self._quit_part2)

    def _quit_part2(self):
        """
        triggered behind self.quit + a short delay
        to let current operations to close cleanly
        """
        if self.state.current_operation in (
            CurrentOperation.BACKUP_IN_PROGRESS,
            CurrentOperation.FORGET_IN_PROGRESS,
        ):
            self.logger.write("Waiting for restic process to be finished")
            QTimer.singleShot(200, self._quit_part2)
        else:
            self.qt_app.quit()



#!/usr/bin/env python3

import getpass
import os
import pwd
import sys

from enacrestic import __version__

# Package related
VERSION_INFO = tuple(map(lambda x: int(x), __version__.split(".")))

PYPI_PROJECT_URL = "https://pypi.org/pypi/enacrestic/json"
UPGRADE_DOC = "https://github.com/EPFL-ENAC/ENACrestic#upgrade"
DEF_CHECK_NEW_VERSION_EVERY_N_DAYS = 7

# App related
DEF_BACKUP_EVERY_N_MINUTES = 30
DEF_FORGET_EVERY_N_BACKUPS = 10

DEF_GUI_AUTOSTART = False

NB_CHRONOS_TO_SAVE = 10

ENACRESTIC_PREF_FOLDER = os.path.expanduser("~/.enacrestic")

RESTIC_USER_PREFS = {
    "FILESFROM": os.path.join(ENACRESTIC_PREF_FOLDER, "bkp_include"),
    "EXCLUDEFILE": os.path.join(ENACRESTIC_PREF_FOLDER, "bkp_exclude"),
    "PASSWORDFILE": os.path.join(ENACRESTIC_PREF_FOLDER, ".pw"),
    "ENV": os.path.join(ENACRESTIC_PREF_FOLDER, "env.sh"),
}
RESTIC_LOGFILE = os.path.join(ENACRESTIC_PREF_FOLDER, "last_backups.log")
RESTIC_CONFFILE = os.path.join(ENACRESTIC_PREF_FOLDER, "prefs.json")
RESTIC_STATEFILE = os.path.join(ENACRESTIC_PREF_FOLDER, "state.json")
PRE_BACKUP_HOOK = os.path.join(ENACRESTIC_PREF_FOLDER, "pre_backup")
RESTIC_AUTOSTART_FILE = os.path.expanduser("~/.config/autostart/enacrestic.desktop")

LOGFILE_ROTATION_EVERY_N_DAYS = 30
LOGFILE_ROTATION_BACKUP_COUNT = 5

ENACRESTIC_BIN = os.path.abspath(sys.argv[0])

USERNAME = getpass.getuser()
UID = pwd.getpwnam(USERNAME).pw_uid
PID_FILE = os.path.join(ENACRESTIC_PREF_FOLDER, "enacrestic.pid")

ICONS_FOLDER = os.path.abspath(f"{__file__}/../pixmaps")

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"



"""
Everything to package enacrestic

Based on:
+ https://packaging.python.org/tutorials/packaging-projects/
+ https://github.com/pypa/sampleproject

Build it with:
$ python3 -m build

This is automated with:
$ make package
"""

import pathlib
import re

from dynaconf import Dynaconf
from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()
pyproject_meta = Dynaconf(
    settings_files=[here / "pyproject.toml"],
)

# ["Samuel Bancal <Samuel.Bancal@epfl.ch>"]
author, author_email = map(
    lambda s: s.strip(),
    re.search(
        r"^([^<]+)<([^>]+)>$", pyproject_meta.get("tool.poetry.authors")[0]
    ).groups(),
)

setup(
    name=pyproject_meta.get("tool.poetry.name"),
    version=pyproject_meta.get("tool.poetry.version"),
    description=pyproject_meta.get("tool.poetry.description"),
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url=pyproject_meta.get("tool.poetry.homepage"),
    author=author,
    author_email=author_email,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: X11 Applications :: Qt",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Archiving :: Backup",
    ],
    keywords="backup, restic",
    package_dir={"": "."},
    packages=["enacrestic"],
    python_requires=">=3.6, <4",
    install_requires=[
        "PyQt5",
        "python-pidfile",
    ],
    extras_require={
        "dev": [],
        "test": [],
    },
    package_data={
        "enacrestic": ["pixmaps/*.png"],
    },
    data_files=[
        ("share/applications", ["enacrestic.desktop"]),
        ("share/icons", ["enacrestic.png"]),
    ],
    entry_points={
        "console_scripts": [
            "enacrestic=enacrestic:main.main",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/EPFL-ENAC/ENACrestic/issues",
        "Source": "https://github.com/EPFL-ENAC/ENACrestic/",
    },
)
